import json
import os
import pickle
import tempfile
import unittest
import sys

from embedded_jubatus import Classifier
from jubatus.classifier.types import EstimateResult
from jubatus.classifier.types import LabeledDatum
from jubatus.common import Datum
import numpy as np


CONFIG = {
    "method": "perceptron",
    "converter": {
        "num_filter_types": {},
        "num_filter_rules": [],
        "string_filter_types": {},
        "string_filter_rules": [],
        "num_types": {},
        "num_rules": [
            {"key": "*", "type": "num"}
        ],
        "string_types": {},
        "string_rules": [
            {"key": "*", "type": "space",
             "sample_weight": "bin", "global_weight": "bin"}
        ]
    },
    "parameter": {}
}


class TestClassifier(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Classifier)
        self.assertRaises(ValueError, Classifier, {})
        self.assertRaises(TypeError, Classifier, {'method': 'hoge'})
        self.assertRaises(RuntimeError, Classifier,
                          {'method': 'hoge', 'converter': {}})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, Classifier, invalid_config)

    def test_num(self):
        x = Classifier(CONFIG)
        self.assertEqual(2, x.train([
            ('Y', Datum({'x': 1})),
            ('N', Datum({'x': -1})),
        ]))

        def _test_classify(x):
            y = x.classify([
                Datum({'x': 1}),
                Datum({'x': -1})
            ])
            self.assertEqual(['Y', 'N'], [list(sorted(
                z, key=lambda x:x.score, reverse=True))[0].label for z in y])
            self.assertEqual(x.get_labels(), {'N': 1, 'Y': 1})

        _test_classify(x)
        model = x.save_bytes()

        self.assertTrue(x.clear())
        self.assertEqual({}, x.get_labels())
        x.set_label('Y')
        x.set_label('N')
        self.assertEqual({'N': 0, 'Y': 0}, x.get_labels())
        x.delete_label(u'Y')
        self.assertEqual({'N': 0}, x.get_labels())

        x = Classifier(CONFIG)
        x.load_bytes(model)
        _test_classify(x)
        self.assertEqual(CONFIG, json.loads(x.get_config()))

        if sys.version_info[0] == 3:
            x = pickle.loads(pickle.dumps(x))
            _test_classify(x)
            self.assertEqual(CONFIG, json.loads(x.get_config()))

    def test_str(self):
        x = Classifier(CONFIG)
        self.assertEqual(2, x.train([
            ('Y', Datum({'x': 'y'})),
            ('N', Datum({'x': 'n'})),
        ]))
        y = x.classify([
            Datum({'x': 'y'}),
            Datum({'x': 'n'})
        ])
        self.assertEqual(['Y', 'N'], [list(sorted(
            z, key=lambda x:x.score, reverse=True))[0].label for z in y])

    def test_types(self):
        x = Classifier(CONFIG)
        x.train([
            LabeledDatum('Y', Datum({'x': 'y'})),
            LabeledDatum('N', Datum({'x': 'n'})),
        ])
        y = x.classify([
            Datum({'x': 'y'}),
            Datum({'x': 'n'})
        ])
        self.assertTrue(isinstance(y[0][0], EstimateResult))
        self.assertEqual(['Y', 'N'], [list(sorted(
            z, key=lambda x:x.score, reverse=True))[0].label for z in y])

    def test_loadsave(self):
        x = Classifier(CONFIG)
        x.train([
            LabeledDatum('Y', Datum({'x': 'y'})),
            LabeledDatum('N', Datum({'x': 'n'})),
        ])
        path = '/tmp/127.0.0.1_0_classifier_hoge.jubatus'

        def _remove_model():
            try:
                os.remove(path)
            except Exception:
                pass

        _remove_model()
        try:
            self.assertEqual({'127.0.0.1_0': '/tmp/127.0.0.1_0_classifier_hoge.jubatus'}, x.save('hoge'))
            self.assertTrue(os.path.isfile(path))
            x = Classifier(CONFIG)
            self.assertTrue(x.load('hoge'))
            y = x.classify([
                Datum({'x': 'y'}),
                Datum({'x': 'n'})
            ])
            self.assertEqual(['Y', 'N'], [list(sorted(
                z, key=lambda x:x.score, reverse=True))[0].label for z in y])
        finally:
            _remove_model()

    def test_numpy(self):
        x = Classifier(CONFIG)
        tdata = np.array([
            [1, 0, 1],
            [0, 1, 1],
        ], dtype='f8')
        ttargets = np.array([1, 0])
        x.partial_fit(tdata, ttargets)
        y = x.predict(np.array([
            [1, 0, 0],
            [0, 1, 0],
        ], dtype='f8'))
        self.assertEqual(1, y[0])
        self.assertEqual(0, y[1])
        y = x.decision_function(np.array([
            [1, 0, 0],
            [0, 1, 0],
        ], dtype='f8'))
        self.assertEqual(2, len(y))
        self.assertTrue(not isinstance(y[0], (list, tuple, np.ndarray)))
        self.assertTrue(y[0] > 0)
        self.assertTrue(y[1] < 0)
        self.assertEqual([0, 1], sorted(x.classes_))

        model = x.save_bytes()
        x = Classifier(CONFIG)
        x.load_bytes(model)
        self.assertEqual([0, 1], sorted(x.classes_))
        y = x.predict(np.array([
            [1, 0, 0],
            [0, 1, 0],
        ], dtype='f8'))
        self.assertEqual(1, y[0])
        self.assertEqual(0, y[1])

    def test_sparse(self):
        from scipy.sparse import csr_matrix

        x = Classifier(CONFIG)
        tdata = csr_matrix(np.array([
            [1, 0, 1],
            [0, 1, 1],
        ], dtype='f8'))
        ttargets = np.array([1, 0])
        x.partial_fit(tdata, ttargets)
        y = x.predict(csr_matrix(np.array([
            [1, 0, 0],
            [0, 1, 0],
        ], dtype='f8')))
        self.assertEqual(1, y[0])
        self.assertEqual(0, y[1])
        y = x.decision_function(csr_matrix(np.array([
            [1, 0, 0],
            [0, 1, 0],
        ], dtype='f8')))
        self.assertEqual(2, len(y))
        self.assertTrue(not isinstance(y[0], (list, tuple, np.ndarray)))
        self.assertTrue(y[0] > 0)
        self.assertTrue(y[1] < 0)

    def test_multilabel(self):
        x = Classifier(CONFIG)
        tdata = np.array([
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0],
        ], dtype='f8')
        ttargets = np.array([1, 2, 0])
        x.partial_fit(tdata, ttargets)
        y = x.predict(np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ], dtype='f8'))
        self.assertTrue(isinstance(y, np.ndarray))
        self.assertEqual(y.dtype, np.int32)
        self.assertTrue(np.equal(y, np.array([0, 1, 2])).all())
        y = x.decision_function(np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ], dtype='f8'))
        self.assertTrue(isinstance(y, np.ndarray))
        self.assertEqual((3, 3), y.shape)
        self.assertTrue(y[0][0] > y[0][1] and y[0][0] > y[0][2])
        self.assertTrue(y[1][1] > y[1][0] and y[1][1] > y[1][2])
        self.assertTrue(y[2][2] > y[2][0] and y[2][2] > y[2][1])

    def test_load_config_from_file(self):
        with tempfile.NamedTemporaryFile() as fd:
            fd.write(json.dumps(CONFIG).encode('utf8'))
            fd.flush()

            x = Classifier(fd.name)
            self.assertEqual(CONFIG, json.loads(x.get_config()))

    def test_load_invalid_config_from_file(self):
        with tempfile.NamedTemporaryFile() as fd:
            fd.write(b'{"hoge')
            fd.flush()
            self.assertRaises(ValueError, Classifier, fd.name)

    def test_issue_30(self):
        path = '/tmp/127.0.0.1_0_classifier_foo.jubatus'
        open(path, 'wb').close()
        try:
            Classifier({'method': 'perceptron', 'converter': {}}).load('foo')
        except Exception:
            pass
        finally:
            os.remove(path)



import json
import random
import unittest

from embedded_jubatus import Bandit
from jubatus.bandit.types import ArmInfo


CONFIG = {
    'method': 'epsilon_greedy',
    'parameter': {
        'assume_unrewarded': False,
        'epsilon': 0.1
    }
}


class TestBandit(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Bandit)
        self.assertRaises(ValueError, Bandit, {})
        self.assertRaises(RuntimeError, Bandit, {'method': 'hoge'})
        invalid_config = dict(CONFIG)
        invalid_config['parameter'] = {'hoge': 0.1}
        self.assertRaises(RuntimeError, Bandit, invalid_config)

    def test(self):
        x = Bandit(CONFIG)

        player = 'player'
        slots = {
            'a': [
                lambda: random.random() < 0.1,
                lambda: random.normalvariate(50, 10),
            ],
            'b': [
                lambda: random.random() < 0.01,
                lambda: random.normalvariate(600, 100),
            ],
            'c': [
                lambda: random.random() < 0.001,
                lambda: random.normalvariate(8000, 1000),
            ],
        }
        keys = list(slots.keys())
        for k in keys:
            self.assertTrue(x.register_arm(k))
        self.assertFalse(x.register_arm(keys[0]))
        self.assertFalse(x.reset(player))

        for _ in range(10):
            arm = x.select_arm(player)
            f0, f1 = slots[arm]
            self.assertTrue(arm in keys)
            x.register_reward(player, arm, f1() if f0() else 0.0)
        info = x.get_arm_info(player)
        self.assertEqual(3, len(info))
        self.assertTrue(isinstance(info[keys[0]], ArmInfo))

        model = x.save_bytes()
        x = Bandit(CONFIG)
        x.load_bytes(model)
        self.assertEqual(CONFIG, json.loads(x.get_config()))
        info = x.get_arm_info(player)
        self.assertEqual(3, len(info))
        self.assertTrue(isinstance(info[keys[0]], ArmInfo))



import json
import unittest

from embedded_jubatus import NearestNeighbor
from jubatus.common import Datum
from jubatus.nearest_neighbor.types import IdWithScore


CONFIG = {
    "method": "euclid_lsh",
    "parameter": {
        "hash_num": 128
    },
    "converter": {
        "num_filter_types": {},
        "num_filter_rules": [],
        "string_filter_types": {},
        "string_filter_rules": [],
        "num_types": {},
        "num_rules": [
            {"key": "*", "type": "num"}
        ],
        "string_types": {},
        "string_rules": [
            {"key": "*", "type": "space",
             "sample_weight": "bin", "global_weight": "bin"}
        ]
    }
}


class TestNearestNeighbor(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, NearestNeighbor)
        self.assertRaises(ValueError, NearestNeighbor, {})
        self.assertRaises(TypeError, NearestNeighbor, {'method': 'hoge'})
        self.assertRaises(RuntimeError, NearestNeighbor,
                          {'method': 'hoge', 'converter': {}})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, NearestNeighbor, invalid_config)

    def test(self):
        x = NearestNeighbor(CONFIG)
        self.assertTrue(x.set_row("a0", Datum({'x': 0})))
        self.assertTrue(x.set_row("a1", Datum({'x': 0.25})))
        self.assertTrue(x.set_row("a2", Datum({'x': 0.5})))
        self.assertTrue(x.set_row("a3", Datum({'x': 1})))
        self.assertTrue(x.set_row("b0", Datum({'x': 10})))
        self.assertTrue(x.set_row("b1", Datum({'x': 10.25})))
        self.assertTrue(x.set_row("b2", Datum({'x': 10.5})))
        self.assertTrue(x.set_row("b3", Datum({'x': 11})))

        def _check_prefix(expected, lst):
            for x in lst:
                self.assertTrue(isinstance(x, IdWithScore))
                self.assertTrue(x.id.startswith(expected))

        ret = x.neighbor_row_from_id("a0", 3)
        self.assertEqual(3, len(ret))
        _check_prefix('a', ret)

        ret = x.neighbor_row_from_datum(Datum({'x': 0.25}), 3)
        self.assertEqual(3, len(ret))
        _check_prefix('a', ret)

        ret = x.similar_row_from_id("b3", 3)
        self.assertEqual(3, len(ret))
        _check_prefix('b', ret)

        ret = x.similar_row_from_datum(Datum({'x': 11}), 3)
        self.assertEqual(3, len(ret))
        _check_prefix('b', ret)

        self.assertEqual(
            set(['a0', 'a1', 'a2', 'a3', 'b0', 'b1', 'b2', 'b3']),
            set(x.get_all_rows()))
        self.assertEqual(CONFIG, json.loads(x.get_config()))
        model = x.save_bytes()

        x = NearestNeighbor(CONFIG)
        x.load_bytes(model)
        self.assertEqual(
            set(['a0', 'a1', 'a2', 'a3', 'b0', 'b1', 'b2', 'b3']),
            set(x.get_all_rows()))



cdef class Stat(_JubatusBase):
    cdef _Stat *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Stat(config)
        self._type, self._model_ver = b'stat', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def push(self, key, value):
        self._handle.push(key.encode('utf8'), value)
        return True

    def sum(self, key):
        return self._handle.sum(key.encode('utf8'))

    def stddev(self, key):
        return self._handle.stddev(key.encode('utf8'))

    def max(self, key):
        return self._handle.max(key.encode('utf8'))

    def min(self, key):
        return self._handle.min(key.encode('utf8'))

    def entropy(self, key):
        return self._handle.entropy()

    def moment(self, key, degree, center):
        return self._handle.moment(key.encode('utf8'), degree, center)



cdef class Graph(_JubatusBase):
    cdef _Graph *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Graph(config)
        self._type, self._model_ver = b'graph', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def create_node(self):
        return self._handle.create_node().decode('ascii')

    def remove_node(self, node_id):
        return self._handle.remove_node(node_id.encode('ascii'))

    def update_node(self, node_id, props):
        cdef prop_t p
        props_py2native(props, p)
        return self._handle.update_node(node_id.encode('ascii'), p)

    def create_edge(self, node_id, e):
        if not isinstance(e, Edge):
            raise ValueError
        if node_id != e.source:
            raise ValueError
        cdef prop_t p
        props_py2native(e.property, p)
        return self._handle.create_edge(node_id.encode('ascii'),
                                        e.target.encode('ascii'), p)

    def update_edge(self, node_id, edge_id, e):
        if not isinstance(e, Edge):
            raise ValueError
        if node_id != e.source:
            raise ValueError
        cdef prop_t p
        props_py2native(e.property, p)
        return self._handle.update_edge(edge_id, p)

    def remove_edge(self, node_id, edge_id):
        self._handle.remove_edge(edge_id)

    def get_centrality(self, node_id, centrality_type, query):
        if not isinstance(query, PresetQuery):
            raise ValueError
        cdef preset_query q
        preset_query_py2native(query, q)
        return self._handle.get_centrality(node_id.encode('ascii'),
                                           centrality_type, q)

    def add_centrality_query(self, query):
        if not isinstance(query, PresetQuery):
            raise ValueError
        cdef preset_query q
        preset_query_py2native(query, q)
        self._handle.add_centrality_query(q)
        return True

    def add_shortest_path_query(self, query):
        if not isinstance(query, PresetQuery):
            raise ValueError
        cdef preset_query q
        preset_query_py2native(query, q)
        self._handle.add_shortest_path_query(q)
        return True

    def remove_centrality_query(self, query):
        if not isinstance(query, PresetQuery):
            raise ValueError
        cdef preset_query q
        preset_query_py2native(query, q)
        self._handle.remove_centrality_query(q)
        return True

    def remove_shortest_path_query(self, query):
        if not isinstance(query, PresetQuery):
            raise ValueError
        cdef preset_query q
        preset_query_py2native(query, q)
        self._handle.remove_shortest_path_query(q)
        return True

    def get_shortest_path(self, query):
        if not isinstance(query, ShortestPathQuery):
            raise ValueError
        cdef preset_query q
        cdef vector[node_id_t] r
        preset_query_py2native(query.query, q)
        r = self._handle.get_shortest_path(query.source.encode('ascii'),
                                           query.target.encode('ascii'),
                                           query.max_hop, q)
        return [str(x) for x in r]

    def update_index(self):
        self._handle.update_index()
        return True

    def get_node(self, node_id):
        cdef node_info n
        n = self._handle.get_node(node_id.encode('ascii'))
        return Node(props_native2py(n.property),
                    edges_native2py(n.in_edges),
                    edges_native2py(n.out_edges))

    def get_edge(self, node_id, edge_id):
        cdef edge_info e
        e = self._handle.get_edge(edge_id)
        return Edge(props_native2py(e.p),
                    str(e.src), str(e.tgt))



import json
import unittest

from embedded_jubatus import Clustering
from jubatus.clustering.types import WeightedDatum
from jubatus.clustering.types import WeightedIndex
from jubatus.clustering.types import IndexedPoint
from jubatus.common import Datum


CONFIG = {
    'method': 'kmeans',
    'parameter': {
        'k': 2,
        'seed': 0,
    },
    'compressor_method': 'simple',
    'compressor_parameter': {
        'bucket_size': 6,
    },
    'distance': 'euclidean',
    'converter': {
        'num_filter_types': {},
        'num_filter_rules': [],
        'string_filter_types': {},
        'string_filter_rules': [],
        'num_types': {},
        'num_rules': [
            {'key': '*', 'type': 'num'}
        ],
        'string_types': {},
        'string_rules': [
            {'key': '*', 'type': 'space',
             'sample_weight': 'bin', 'global_weight': 'bin'}
        ]
    }
}


class TestClustering(unittest.TestCase):
    def test_invalid_configs(self):
        self.assertRaises(TypeError, Clustering)
        self.assertRaises(Exception, Clustering, {})
        self.assertRaises(Exception, Clustering, {'method': 'hoge'})
        self.assertRaises(Exception, Clustering,
                          {'method': 'hoge', 'converter': {}})
        self.assertRaises(Exception, Clustering,
                          {'method': 'hoge', 'converter': {},
                           'compressor_method': 'hoge'})
        self.assertRaises(RuntimeError, Clustering,
                          {'method': 'hoge', 'converter': {},
                           'compressor_method': 'hoge',
                           'compressor_parameter': {}})
        self.assertRaises(RuntimeError, Clustering,
                          {'method': 'hoge', 'converter': {},
                           'compressor_method': 'hoge',
                           'compressor_parameter': {},
                           'distance': 'hoge'})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, Clustering, invalid_config)

    def test_default_configs(self):
        omitted_config = dict(CONFIG)
        del omitted_config['distance']
        try:
            x = Clustering(omitted_config)
        except:
            self.fail()
        self.assertTrue(True)

    def test(self):
        x = Clustering(CONFIG)
        self.assertEqual(0, x.get_revision())
        self.assertTrue(x.push([
            IndexedPoint('a', Datum({'x': 1.0})),
            IndexedPoint('b', Datum({'x': 0.9})),
            IndexedPoint('c', Datum({'x': 1.1})),
            IndexedPoint('d', Datum({'x': 5.0})),
            IndexedPoint('e', Datum({'x': 4.9})),
            IndexedPoint('f', Datum({'x': 5.1})),
        ]))
        self.assertEqual(1, x.get_revision())
        centers = x.get_k_center()
        self.assertTrue(isinstance(centers, list))
        self.assertEqual(2, len(centers))
        self.assertTrue(isinstance(centers[0], Datum))
        members = x.get_core_members()
        self.assertTrue(isinstance(members, list))
        self.assertEqual(2, len(members))
        self.assertTrue(isinstance(members[0], list))
        self.assertTrue(isinstance(members[0][0], WeightedDatum))
        c = x.get_nearest_center(Datum({'x': 1.05}))
        self.assertTrue(isinstance(c, Datum))
        self.assertTrue(c.num_values[0][1] >= 0.9 and
                        c.num_values[0][1] <= 1.1)
        c = x.get_nearest_members(Datum({'x': 1.05}))
        self.assertTrue(isinstance(c, list))
        self.assertTrue(isinstance(c[0], WeightedDatum))

        c = x.get_core_members_light()
        self.assertTrue(isinstance(c, list))
        self.assertTrue(isinstance(c[0], list))
        self.assertTrue(isinstance(c[0][0], WeightedIndex))

        c = x.get_nearest_members_light(Datum({'x': 1.05}))
        self.assertTrue(isinstance(c, list))
        self.assertTrue(isinstance(c[0], WeightedIndex))

        model = x.save_bytes()
        x = Clustering(CONFIG)
        x.load_bytes(model)

        self.assertEqual(CONFIG, json.loads(x.get_config()))
        self.assertEqual(1, x.get_revision())
        self.assertEqual(len(centers), len(x.get_k_center()))



import json
import unittest

from embedded_jubatus import Graph
from jubatus.graph.types import Edge
from jubatus.graph.types import Node
from jubatus.graph.types import PresetQuery
from jubatus.graph.types import ShortestPathQuery


CONFIG = {
    "parameter": {
        "damping_factor": 0.9,
        "landmark_num": 5
    },
    "method": "graph_wo_index"
}


class TestGraph(unittest.TestCase):
    def test(self):
        g = Graph(CONFIG)
        self.assertTrue(g.add_shortest_path_query(PresetQuery([], [])))

        n0 = g.create_node()
        n1 = g.create_node()
        self.assertTrue(g.remove_node(g.create_node()))
        self.assertEqual('0', n0)
        self.assertEqual('1', n1)
        self.assertTrue(g.update_node(n0, {'hoge': 'piyo'}))

        e0 = g.create_edge(n0, Edge({'foo': 'bar'}, n0, n1))
        self.assertEqual(e0, 3)

        ni = g.get_node(n0)
        self.assertTrue(isinstance(ni, Node))
        self.assertEqual({'hoge': 'piyo'}, ni.property)
        self.assertEqual([], ni.in_edges)
        self.assertEqual([e0], ni.out_edges)
        ei = g.get_edge(n0, e0)
        self.assertTrue(isinstance(ei, Edge))
        self.assertEqual({'foo': 'bar'}, ei.property)
        self.assertEqual(n0, ei.source)
        self.assertEqual(n1, ei.target)
        self.assertTrue(g.update_edge(n0, e0, Edge({'foo': 'bar2'}, n0, n1)))
        ei = g.get_edge(n0, e0)
        self.assertEqual({'foo': 'bar2'}, ei.property)

        self.assertTrue(g.add_centrality_query(PresetQuery([], [])))
        self.assertTrue(g.update_index())
        self.assertTrue(isinstance(
            g.get_centrality(n0, 0, PresetQuery([], [])), float))
        self.assertTrue(g.remove_centrality_query(PresetQuery([], [])))

        self.assertEqual([n0, n1], g.get_shortest_path(
            ShortestPathQuery(n0, n1, 100, PresetQuery([], []))))
        model = g.save_bytes()
        self.assertTrue(g.remove_shortest_path_query(PresetQuery([], [])))

        g = Graph(CONFIG)
        g.load_bytes(model)
        self.assertEqual(CONFIG, json.loads(g.get_config()))
        self.assertEqual([n0, n1], g.get_shortest_path(
            ShortestPathQuery(n0, n1, 100, PresetQuery([], []))))
        self.assertEqual('4', g.create_node())



from libc.stdint cimport uint64_t
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.map cimport map
from cython.operator cimport dereference
from cython.operator cimport preincrement

cimport numpy as c_np

from _wrapper cimport _Anomaly
from _wrapper cimport _Bandit
from _wrapper cimport _Burst
from _wrapper cimport _Classifier
from _wrapper cimport _Clustering
from _wrapper cimport _NearestNeighbor
from _wrapper cimport _Recommender
from _wrapper cimport _Regression
from _wrapper cimport _Stat
from _wrapper cimport _Weight
from _wrapper cimport _Graph
from _wrapper cimport arm_info
from _wrapper cimport classify_result_elem
from _wrapper cimport datum
from _wrapper cimport keyword_params
from _wrapper cimport keyword_with_params
from _wrapper cimport lexical_cast
from _wrapper cimport sfv_t
from _wrapper cimport prop_t
from _wrapper cimport node_id_t
from _wrapper cimport edge_id_t
from _wrapper cimport node_info
from _wrapper cimport edge_info
from _wrapper cimport preset_query
from _wrapper cimport indexed_point

from jubatus.anomaly.types import IdWithScore as AnomalyIdWithScore
from jubatus.bandit.types import ArmInfo
from jubatus.burst.types import Batch
from jubatus.burst.types import Document
from jubatus.burst.types import KeywordWithParams
from jubatus.burst.types import Window
from jubatus.classifier.types import EstimateResult
from jubatus.classifier.types import LabeledDatum
from jubatus.clustering.types import WeightedDatum
from jubatus.clustering.types import WeightedIndex
from jubatus.common.datum import Datum
from jubatus.nearest_neighbor.types import IdWithScore as NNIdWithScore
from jubatus.recommender.types import IdWithScore as RecommenderIdWithScore
from jubatus.regression.types import ScoredDatum
from jubatus.weight.types import Feature
from jubatus.graph.types import Edge
from jubatus.graph.types import Node
from jubatus.graph.types import PresetQuery
from jubatus.graph.types import Query
from jubatus.graph.types import ShortestPathQuery


cdef class _JubatusBase:
    cdef object _type
    cdef int _model_ver

    def __init__(self, config):
        import json
        if isinstance(config, str):
            # loads config from file
            config = open(config, 'rb').read().decode('utf8')
            # JSON parse test
            json.loads(config)
        else:
            config = json.dumps(config, sort_keys=True, indent=4)
        self._init(config.encode('utf8'))

    def _get_model_path(self, id_):
        host, port, typ = '127.0.0.1', 0, self._type
        if str != bytes and isinstance(typ, bytes):
            typ = typ.decode('ascii')
        path = '/tmp/{host}_{port}_{type}_{id}.jubatus'.format(
            host=host, port=port, type=typ, id=id_)
        return (path, '{host}_{port}'.format(host=host, port=port))

    def load(self, id_):
        path, name = self._get_model_path(id_)
        with open(path, 'rb') as f:
            self.load_bytes(f.read())
        return True

    def save(self, id_):
        path, name = self._get_model_path(id_)
        with open(path, 'wb') as f:
            f.write(self.save_bytes())
        return {name: path}

    def get_status(self):
        raise RuntimeError

    def do_mix(self):
        raise RuntimeError

    def get_proxy_status(self):
        raise RuntimeError

    def get_name(self):
        raise RuntimeError

    def set_name(self, new_name):
        raise RuntimeError

    def get_client(self):
        raise RuntimeError

    def __getstate__(self):
        return (self.get_config(), self.save_bytes())

    def __setstate__(self, state):
        import json
        cfg, model = state
        self.__init__(json.loads(cfg))
        self.load_bytes(model)

include 'types.pyx'
include 'anomaly.pyx'
include 'bandit.pyx'
include 'burst.pyx'
include 'classifier.pyx'
include 'clustering.pyx'
include 'nearest_neighbor.pyx'
include 'recommender.pyx'
include 'regression.pyx'
include 'stat.pyx'
include 'weight.pyx'
include 'graph.pyx'



import json
import unittest

from embedded_jubatus import Regression
from jubatus.common import Datum
from jubatus.regression.types import ScoredDatum
import numpy as np


CONFIG = {
    "method": "PA1",
    "parameter": {
        "sensitivity": 0.1,
        "regularization_weight": 3.402823e+38
    },
    "converter": {
        "num_filter_types": {},
        "num_filter_rules": [],
        "string_filter_types": {},
        "string_filter_rules": [],
        "num_types": {},
        "num_rules": [
            {"key": "*", "type": "num"}
        ],
        "string_types": {},
        "string_rules": [
            {"key": "*", "type": "space",
             "sample_weight": "bin", "global_weight": "bin"}
        ]
    }
}


class TestRegression(unittest.TestCase):
    def test_invalid_config(self):
        self.assertRaises(TypeError, Regression)
        self.assertRaises(ValueError, Regression, {})
        self.assertRaises(TypeError, Regression, {'method': 'hoge'})
        self.assertRaises(RuntimeError, Regression,
                          {'method': 'hoge', 'converter': {}})
        invalid_config = dict(CONFIG)
        invalid_config['method'] = 'hoge'
        self.assertRaises(RuntimeError, Regression, invalid_config)

    def test(self):
        x = Regression(CONFIG)
        self.assertEqual(5, x.train([
            ScoredDatum(0.0, Datum({'x': 1.0})),
            ScoredDatum(1.0, Datum({'x': 2.0})),
            ScoredDatum(2.0, Datum({'x': 4.0})),
            ScoredDatum(3.0, Datum({'x': 8.0})),
            ScoredDatum(4.0, Datum({'x': 16.0})),
        ]))
        ret = x.estimate([
            Datum({'x': 32.0}),
            Datum({'x': 1.5}),
        ])
        self.assertEqual(2, len(ret))
        self.assertTrue(ret[0] >= 8.0 and ret[0] < 9.0)
        self.assertTrue(ret[1] >= 0.0 and ret[1] < 1.0)
        self.assertEqual(CONFIG, json.loads(x.get_config()))

        model = x.save_bytes()
        x = Regression(CONFIG)
        x.load_bytes(model)
        self.assertEqual(ret, x.estimate([
            Datum({'x': 32.0}),
            Datum({'x': 1.5}),
        ]))

    def test_numpy(self):
        X = np.array([[1.0], [2.0], [4.0], [8.0], [16.0]])
        y = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
        reg = Regression(CONFIG)
        reg.fit(X, y)
        ret = reg.predict(np.array([[32.0], [1.5]]))
        self.assertIsInstance(ret, np.ndarray)
        self.assertTrue(ret[0] >= 8.0 and ret[0] < 9.0)
        self.assertTrue(ret[1] >= 0.0 and ret[1] < 1.0)



cdef class Weight(_JubatusBase):
    cdef _Weight *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Weight(config)
        self._type, self._model_ver = b'weight', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def update(self, data):
        cdef datum d
        cdef sfv_t r
        datum_py2native(data, d)
        r = self._handle.update(d)
        return _to_features(r)

    def calc_weight(self, data):
        cdef datum d
        cdef sfv_t r
        datum_py2native(data, d)
        r = self._handle.calc_weight(d)
        return _to_features(r)

cdef _to_features(sfv_t& r):
    cdef int sz
    sz = r.size()
    ret = [None] * sz
    for i in range(sz):
        ret[i] = Feature(r[i].first.decode('utf8'),
                         r[i].second)
    return ret



cdef class Regression(_JubatusBase):
    cdef _Regression *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Regression(config)
        self._type, self._model_ver = b'regression', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def train(self, data):
        cdef datum d
        if isinstance(data, list):
            if len(data) == 0:
                return 0
            if isinstance(data[0], ScoredDatum):
                for r in data:
                    score = float(r.score)
                    datum_py2native(r.data, d)
                    self._handle.train(score, d)
            elif isinstance(data[0], (tuple, list)):
                for score, datum in data:
                    datum_py2native(datum, d)
                    self._handle.train(float(score), d)
            else:
                raise ValueError
            return len(data)
        else:
            raise ValueError

    def estimate(self, data):
        cdef datum d
        if isinstance(data, list):
            ret = []
            for datum in data:
                datum_py2native(datum, d)
                ret.append(self._handle.estimate(d))
            return ret
        else:
            raise ValueError

    def fit(self, X, y):
        self.clear()
        return self.partial_fit(X, y)

    def partial_fit(self, X, y):
        import numpy as np
        if len(X.shape) != 2 or len(y.shape) != 1 or X.shape[0] != y.shape[0]:
            raise ValueError('invalid shape')
        cdef int i
        cdef double score
        cdef rows = X.shape[0]
        cdef datum d
        cdef vector[string] cache
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        if not (is_ndarray or is_csr):
            raise ValueError
        for i in range(rows):
            if is_ndarray:
                ndarray_to_datum(X, i, d, cache)
            else:
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
            score = y[i]
            self._handle.train(score, d)
        return self

    def predict(self, X):
        import numpy as np
        if len(X.shape) != 2:
            raise ValueError('invalid shape')
        cdef int i
        cdef rows = X.shape[0]
        cdef datum d
        cdef vector[string] cache
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        if not (is_ndarray or is_csr):
            raise ValueError
        ret = np.zeros([rows], dtype=np.float32)
        for i in range(rows):
            if is_ndarray:
                ndarray_to_datum(X, i, d, cache)
            else:
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
            ret[i] = self._handle.estimate(d)
        return ret



import json
import math
import unittest

from embedded_jubatus import Weight
from jubatus.common import Datum
from jubatus.weight.types import Feature


CONFIG = {
    "converter": {
        "num_filter_types": {},
        "num_filter_rules": [],
        "string_filter_types": {},
        "string_filter_rules": [],
        "num_types": {},
        "num_rules": [
            {"key": "n0", "type": "num"},
            {"key": "n1", "type": "log"},
            {"key": "n2", "type": "str"}
        ],
        "string_types": {},
        "string_rules": [
            {"key": "t0", "type": "space",
             "sample_weight": "bin", "global_weight": "bin"},
            {"key": "t1", "type": "space",
             "sample_weight": "tf", "global_weight": "idf1"}
        ]
    }
}


class TestWeight(unittest.TestCase):
    def test(self):
        w = Weight(CONFIG)
        d = Datum({'n0': 1, 'n1': 2, 'n2': 3, 't0': 'hello world'})
        for r in [w.update(d), w.calc_weight(d)]:
            self.assertEqual(5, len(r))
            for x in r:
                self.assertTrue(isinstance(x, Feature))
            m = dict([(x.key, x.value) for x in r])
            self.assertEqual(5, len(m))
            self.assertEqual(1.0, m['n0@num'])
            self.assertAlmostEqual(math.log(2), m['n1@log'])
            self.assertEqual(1.0, m['n2@str$3'])

        w.update(Datum({'t1': 'hello world'}))
        w.update(Datum({'t1': 'foo bar'}))
        w.update(Datum({'t1': 'hello'}))
        d = Datum({'t1': 'foo bar hello world hoge'})
        r0 = dict([(x.key, x.value) for x in w.calc_weight(d)])

        model = w.save_bytes()
        w = Weight(CONFIG)
        w.load_bytes(model)
        self.assertEqual(CONFIG, json.loads(w.get_config()))
        r1 = dict([(x.key, x.value) for x in w.calc_weight(d)])
        self.assertEqual(r0, r1)



cdef _to_window(const pair[double, vector[_Burst.Batch]]& r):
    return Window(r.first, [
        Batch(r.second[i].all_data_count,
              r.second[i].relevant_data_count,
              r.second[i].burst_weight)
        for i in range(r.second.size())
    ])

cdef _to_window_list(map[string, pair[double, vector[_Burst.Batch]]]& r):
    return {it.first.decode('utf8'): _to_window(it.second) for it in r}

cdef class Burst(_JubatusBase):
    cdef _Burst *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Burst(config)
        self._type, self._model_ver = b'burst', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def add_documents(self, data):
        ret = 0
        for doc in data:
            if self._handle.add_document(doc.text.encode('utf8'), doc.pos):
                ret += 1
        if ret > 0:
            self._handle.calculate_results()
        return ret

    def get_result(self, keyword):
        return _to_window(self._handle.get_result(keyword.encode('utf8')))

    def get_result_at(self, keyword, pos):
        return _to_window(self._handle.get_result_at(keyword.encode('utf8'), pos))

    def get_all_bursted_results(self):
        cdef map[string, pair[double, vector[_Burst.Batch]]] r
        r = self._handle.get_all_bursted_results()
        return _to_window_list(r)

    def get_all_bursted_results_at(self, pos):
        cdef map[string, pair[double, vector[_Burst.Batch]]] r
        r = self._handle.get_all_bursted_results_at(pos)
        return _to_window_list(r)

    def get_all_keywords(self):
        cdef vector[keyword_with_params] r
        r = self._handle.get_all_keywords()
        return [
            KeywordWithParams(r[i].keyword.decode('utf8'), r[i].scaling_param, r[i].gamma)
            for i in range(r.size())
        ]

    def add_keyword(self, keyword):
        cdef keyword_params kp
        kp.scaling_param = keyword.scaling_param
        kp.gamma = keyword.gamma
        return self._handle.add_keyword(keyword.keyword.encode('utf8'), kp)

    def remove_keyword(self, keyword):
        return self._handle.remove_keyword(keyword.encode('utf8'))

    def remove_all_keywords(self):
        return self._handle.remove_all_keywords()



cdef class NearestNeighbor(_JubatusBase):
    cdef _NearestNeighbor *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _NearestNeighbor(config)
        self._type, self._model_ver = b'nearest_neighbor', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def set_row(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        self._handle.set_row(id_.encode('utf8'), d)
        return True

    def neighbor_row_from_id(self, id_, size):
        cdef vector[pair[string, float]] ret
        ret = self._handle.neighbor_row_from_id(id_.encode('utf8'), size)
        return [
            NNIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def neighbor_row_from_datum(self, dat, size):
        cdef datum d
        cdef vector[pair[string, float]] ret
        datum_py2native(dat, d)
        ret = self._handle.neighbor_row_from_datum(d, size)
        return [
            NNIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def similar_row_from_id(self, id_, size):
        cdef vector[pair[string, float]] ret
        ret = self._handle.similar_row_from_id(id_.encode('utf8'), size)
        return [
            NNIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def similar_row_from_datum(self, dat, size):
        cdef datum d
        cdef vector[pair[string, float]] ret
        datum_py2native(dat, d)
        ret = self._handle.similar_row_from_datum(d, size)
        return [
            NNIdWithScore(ret[i].first.decode('utf8'), ret[i].second)
            for i in range(ret.size())
        ]

    def get_all_rows(self):
        cdef vector[string] ret = self._handle.get_all_rows()
        return [(<string>ret[i]).decode('utf8') for i in range(ret.size())]



import os.path
import platform

from setuptools import Extension
from setuptools import setup


def read(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


def _setup():
    include_dirs = []
    install_requires = read('requirements.txt').split('\n')
    build_requires = [
        line for line in install_requires
        if line.startswith('numpy') or line.startswith('Cython')]

    try:
        from Cython.Build import cythonize
    except ImportError:
        cythonize = None
    try:
        import numpy
        include_dirs.append(numpy.get_include())
    except ImportError:
        pass

    if platform.system() == 'Darwin':
        extra_compile_args = ['-std=c++11']
    else:
        extra_compile_args = []

    kwargs = dict(
        name='embedded_jubatus',
        version=read('VERSION').rstrip(),
        description='embedded-jubatus-python is a Python bridge to call Jubatus Core library.',
        long_description=read('README.rst'),
        author='PFN & NTT',
        author_email='jubatus-team@googlegroups.com',
        url='http://jubat.us',
        download_url='http://pypi.python.org/pypi/embedded_jubatus',
        license='LGPLv2',
        classifiers=[
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
        ],
        build_requires=build_requires,
        install_requires=install_requires,
        test_suite='tests',
    )
    if cythonize:
        kwargs['ext_modules'] = cythonize([
            Extension(
                'embedded_jubatus',
                [
                    'src/embedded_jubatus.pyx',
                    'src/_wrapper.cpp',
                    'src/_model.cpp'
                ],
                include_dirs=include_dirs,
                libraries=['jubatus_core', 'jubatus_util_text'],
                language='c++',
                extra_compile_args=extra_compile_args)
        ])
    setup(**kwargs)


if __name__ == '__main__':
    _setup()



cdef class Clustering(_JubatusBase):
    cdef _Clustering *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Clustering(config)
        self._type, self._model_ver = b'clustering', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def push(self, points):
        cdef datum d
        cdef vector[indexed_point] v
        cdef indexed_point ip
        for p in points:
            datum_py2native(p.point, d)
            ip.id = p.id.encode('utf8')
            ip.point = d
            v.push_back(ip)
        self._handle.push(v)
        return True

    def get_revision(self):
        return self._handle.get_revision()

    def get_core_members(self):
        cdef vector[vector[pair[double, datum]]] r
        r = self._handle.get_core_members()
        ret = []
        for i in range(r.size()):
            ret.append([
                WeightedDatum(r[i][j].first, datum_native2py(r[i][j].second))
                for j in range(r[i].size())
            ])
        return ret

    def get_k_center(self):
        cdef vector[datum] r = self._handle.get_k_center()
        return [datum_native2py(r[i]) for i in range(r.size())]

    def get_nearest_center(self, point):
        cdef datum d
        cdef datum r
        datum_py2native(point, d)
        r = self._handle.get_nearest_center(d)
        return datum_native2py(r)

    def get_nearest_members(self, point):
        cdef datum d
        cdef vector[pair[double, datum]] r
        datum_py2native(point, d)
        r = self._handle.get_nearest_members(d)
        return [
            WeightedDatum(r[i].first, datum_native2py(r[i].second))
            for i in range(r.size())
        ]

    def get_core_members_light(self):
        cdef vector[vector[pair[double, string]]] r
        r = self._handle.get_core_members_light()
        ret = []
        for c in r:
            tmp = []
            for m in c:
                tmp.append(WeightedIndex(m.first, m.second))
            ret.append(tmp)
        return ret

    def get_nearest_members_light(self, point):
        cdef vector[pair[double, string]] r
        cdef datum d
        datum_py2native(point, d)
        r = self._handle.get_nearest_members_light(d)
        ret = []
        for m in r:
            ret.append(WeightedIndex(m.first, m.second))
        return ret



cdef class Anomaly(_JubatusBase):
    cdef _Anomaly *_handle

    def __cinit__(self):
        self._handle = NULL

    def __dealloc__(self):
        if self._handle != NULL:
            del self._handle

    def _init(self, config):
        self._handle = new _Anomaly(config)
        self._type, self._model_ver = b'anomaly', 1

    def get_config(self):
        return self._handle.get_config().decode('utf8')

    def save_bytes(self):
        return self._handle.dump(self._type, self._model_ver)

    def load_bytes(self, x):
        return self._handle.load(x, self._type, self._model_ver)

    def clear(self):
        self._handle.clear()
        return True

    def clear_row(self, id_):
        self._handle.clear_row(id_.encode('utf8'))
        return True

    def add(self, row):
        cdef datum d
        cdef pair[string, float] r
        datum_py2native(row, d)
        r = self._handle.add(d)
        return AnomalyIdWithScore(r.first.decode('utf8'), r.second)

    def add_bulk(self, data):
        cdef datum d
        cdef vector[pair[string, datum]] vec
        cdef vector[string] ret
        for x in data:
            datum_py2native(x, d)
            vec.push_back(pair[string, datum](self._handle.get_next_id(), d))
        ret = self._handle.add_bulk(vec)
        return [i.decode('ascii') for i in ret]

    def update(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.update(id_.encode('utf8'), d)

    def overwrite(self, id_, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.overwrite(id_.encode('utf8'), d)

    def calc_score(self, row):
        cdef datum d
        datum_py2native(row, d)
        return self._handle.calc_score(d)

    def get_all_rows(self):
        cdef vector[string] ret = self._handle.get_all_rows()
        return [i.decode('ascii') for i in ret]

    def fit(self, X):
        self.partial_fit(X)

    def partial_fit(self, X):
        import numpy as np
        cdef datum d
        cdef vector[string] cache
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        cdef rows
        cdef vector[pair[string, datum]] vec
        if not (is_ndarray or is_csr):
            raise ValueError
        rows = X.shape[0]
        if is_ndarray:
            for i in range(rows):
                ndarray_to_datum(X, i, d, cache)
                vec.push_back(pair[string, datum](self._handle.get_next_id(), d))
        else:
            for i in range(rows):
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
                vec.push_back(pair[string, datum](self._handle.get_next_id(), d))
        self._handle.add_bulk(vec)

    def decision_function(self, X):
        import numpy as np
        cdef datum d
        cdef vector[string] cache
        cdef int is_ndarray = isinstance(X, np.ndarray)
        cdef int is_csr = (type(X).__name__ == 'csr_matrix')
        cdef rows
        if not (is_ndarray or is_csr):
            raise ValueError
        rows = X.shape[0]
        y = np.zeros((rows,))
        if is_ndarray:
            for i in range(rows):
                ndarray_to_datum(X, i, d, cache)
                y[i] = self._handle.calc_score(d)
        else:
            for i in range(rows):
                csr_to_datum(X.data, X.indices, X.indptr, i, d, cache)
                y[i] = self._handle.calc_score(d)
        return y
